#!/usr/local/bin/perl -wT
#
# $Id: mserv_irman.pl,v 1.23 2004/08/19 22:48:54 abs Exp $
# 
#	Copyright (c) 2002, 2003, 2004 David Brownlee <abs@netbsd.org>.
#	All rights reserved. Provided as-is without express or implied
#	warranties.
#
#	Redistribution and use in source and binary forms, with or without
#	modification, are permitted provided the above copyright and this
#	notice is retained.

my $version = '0.20';
my $SYSCONFDIR = '/usr/local/etc';

=head1 NAME

mserv_irman - control mserv mp3 jukebox program via infra-red remote controls.

=head1 SYNOPSIS

mserv_irman [-v] [-c conf]

=head1 DESCRIPTION

mserv is an mp3 jukebox who's features include allowing people to
rate songs so it can pick tracks appropriate to who is currently
connected.

mserv_irman extends this to remote controls using irman receivers from
evation.com to control one or more mserv jukeboxes.

It will automatically connect to the mserv to ensure a user's ratings are
active when they use their remote control.

=head1 INSTALLATION

If running NetBSD or pkgsrc on another system, install
pkgsrc/audio/mserv_irman, otherwise follow the (brief) install notes in the
README.

=head1 CONFIGURATION

It is recommended you configure mserv first, then the irman, and finally
mserv_irman to hook them together.

=head2 mserv configuration

There are various other options when setting up mserv, this is intended to
provide a 'quick setup' route.

=over 4

=item -

Create an mserv user (eg: 'mserv').

=item -

As that user run 'mserv' then ^C.

=item -

Edit ~mserv/.mserv/config to point path_tracks to your mp3s (or put all the mp3s in ~mserv/.mserv/tracks).

=item -

Login to the mserv (telnet localhost 4444), and create a user for each person
who will be using the system. 'create <user> <pass> user'. You will need the
usernames and passwords when you setup the mserv_irman.conf later.

=back

=head2 irman setup

If you have not installed from pkgsrc ensure you have installed irman_name
(modified test_name from libirman - see README).

=over 4

=item -

Connect up the irman to a serial port and ensure the 'port' is set correctly
in etc/irman.conf.

=item -

chown the serial device to the user who will run mserv_irman (usually mserv).

=item -

Run irman_name - it should report 'ok' if it can initialise the irman.

=item -

Point a remote control anywhere near the irman and press some buttons. It
should report 'rx ....' for each button.

=back

=head2 irman remote configuration

Now you are ready to configure in each remote. Repeat the following for
each remote.

=over 4

=item -

Press each button briefly in turn on a remote and note the rx number.
Note that some remote controls may repeat buttons.

=item -

For each key add a line to etc/irman.conf of the form:

    bind <remotename>-<keyname>	<rx_number>

Where <remotename> is the (user specified) name of the remote
control unit which can contain alphanumeric and '-' characters,
<keyname> is a key from the recognised list (see RECOGNISED KEYS
below).

A sample section for a Hauppauge WinTV PVR would be:

    bind hauppauge-wintvpvr-tv              55a000000000
    bind hauppauge-wintvpvr-ch+             575000000000
    bind hauppauge-wintvpvr-radio           559000000000
    bind hauppauge-wintvpvr-vol-            544000000000
    bind hauppauge-wintvpvr-fullscreen      57b000000000
    bind hauppauge-wintvpvr-vol+            545000000000
    bind hauppauge-wintvpvr-mute            558000000000
    bind hauppauge-wintvpvr-ch-             574000000000
    bind hauppauge-wintvpvr-source          577000000000
    bind hauppauge-wintvpvr-1               554000000000
    bind hauppauge-wintvpvr-2               557000000000
    bind hauppauge-wintvpvr-3               556000000000
    bind hauppauge-wintvpvr-4               551000000000
    bind hauppauge-wintvpvr-5               550000000000
    bind hauppauge-wintvpvr-6               553000000000
    bind hauppauge-wintvpvr-7               552000000000
    bind hauppauge-wintvpvr-8               55d000000000
    bind hauppauge-wintvpvr-9               55c000000000
    bind hauppauge-wintvpvr-pause           54b000000000
    bind hauppauge-wintvpvr-repeat          555000000000
    bind hauppauge-wintvpvr-next            573000000000

If you wish to map more than one button on a given control to
the same feature you should append '.n' to the second and subsequent
buttons. Eg 'vol+' 'vol+.2' 'vol+.3' etc.

=item -

Kill and restart irman_name and confirm it now reports the correct
names when each button is pressed.

=back

=head2 mserv_irman configuration

Finally you need ${SYSCONFDIR}/mserv_irman.conf. It contains four sections:

An (optional) general section which permits you to tune overall
configuration.

    [general]

    ### Ignore any repeated key for this number of seconds
    # default_debounce = 0.5

    ### If an owner is idle for the nyumber of seconds, disconnect them
    # idle_timeout = 3600

    ### Insert this many favourite tracks
    # favourite_inserts = 5

    ### Check sockets, plus check owner idle. Should not need to change.
    # flush_timeout = 60

    ### Optional percentage increase/decrease for volume control
    # volume_down = 10
    # volume_up = 10

An (optional) irman section which links irman receivers to mserv servers.
mserv_irman can handle multiple irman receivers and multiple mserv servers.
The system will default to a single irman on the default serial port linked
to an mserv on the default port on the current machine (localhost:4444).

    ### List of <command to recieve irman data> = <host:port of mserv>
    ### Leave empty to default to 'irman_name' and 'localhost:4444'
    [irman] 

A list of valid 'owners', each user and pass values from mserv.

    ### Add list of owners here, in the form '<mservuser> = <mservpass>'
    [owners] 

A list of valid remotes, linking each remote to a user. A given user can
have more than one remote.

    ### Add your remotes here, in the form '<remotename> = <mservuser>'
    [remotes]

Each section can be overridden on a per host basis by adding _`hostname -s` to
the section key - for example [irman_iris]

=head1 RECOGNISED KEYS

    vol-, vol+	    Volume down and up
    mute	    Volume to 10%
    play	    Play music
    pause/10	    Pause or resume - also any unrecognised key
    stop	    Stop playing (and disconnect from mserv)
    next/fwd/ch+    Skip current track
    prev/rew/ch-    Put previous track on queue, then skip current track
    repeat	    Put current track back onto queue to play again
    recall	    Put previous track on the queue to play again
    1, 2, 3, 4, 5   Rate track 1=AWFUL 2=BAD 3=NEUTRAL 4=GOOD 5=SUPERB
    6		    Add five unrated tracks to the queue (*1)
    7		    Add five tracks from current album (*2)
    8		    Add five tracks from current artist (*2)
    9		    Add five random SUPERB tracks to queue (*3)
    0/reset/power/cancel/clear    Clear queue

	(*1) Will try to queue tracks that no-one connected has rated, and
	     fall back to the current user.
	(*2) Tracks rated AWFUL/BAD/NEUTRAL will not be picked, will try to
	     queue UNRATED/UNHEARD tracks before SUPERB/GOOD.
	(*3) Will try to queue tracks that have not been rated BAD/AWFUL by
	     other connected users first.

=head1 SETTING UP MULTIPLE MSERV SERVERS

This is a more advanced section for those who wish to run multiple mserv
servers (One each for the office, bedroom, and bathroom maybe?)

=head2 mserv

Assuming you want to share user passwords and track ratings between
more than one simultaneous mserv sessions.

=over 4

=item -

From the mserv user's home directory execute:

    mkdir .mserv2
    ln -sf ../.mserv/acl .mserv2/acl
    ln -sf ../.mserv/trackinfo .mserv2/trackinfo
    ln -sf ../.mserv/webacl .mserv2/webacl
    cp ../.mserv/config .mserv2/config

=item -

Edit .mserv2/config and update the audio device.

=item -

Run the second mserv with a new root dir portnumber: 

    mserv -r .mserv2 -p 4445 &

=back

mserv will automatically merge in any ratings update so it is safe to
rate the same song via different mserv servers. It will however have
issues merging in new users, so when adding a user shut down all but one
mserv, create the user on that mserv, then start up the other mservs.

=head2 mserv_irman

Update the [irman] section in mserv_irman.conf to list each irman and its
destination mserv. If you are running an mserv on a remote system you may
want to setup the appropriate ssh keys and have an entry something like:

    ssh -1 shim 'pkill irman_name ; irman_name' = shim:4444

=head1 CAVEATS

Where to begin... here are a few of my favourites:

=over 4

=item -

The mserv_irman.conf file contains mserv passwords in the clear. It should
only be readable by the user running mserv_irman. mserv_irman should also
enforce this.

=item -

If an mserv to which mserv_irman has an open connection dies it will abort
next time it tries to write to that socket.

=item -

If an irman connection dies mserv_irman will close it down and (if there
are any remaining active) continue running. It should periodically try to
restart any closed irman connections.

=item -

Debouncing is completely broken on slower machines. We need to get a timestamp
for each button to make this work correctly. Not difficult, just a little
time.

=item -

When queuing a set of songs it should weight the selection by the rating,
rather then just cutting off NEUTRAL, BAD, and AWFUL.

=item -

There should be a clean way to hook in non mserv uses of other buttons.

=back


=cut

use strict;
use Getopt::Std;
use IO::Socket::INET;
use IO::Select;
use IO::File;
use Config::IniFiles;

my (%opt, %cfg);

$ENV{PATH} = '/usr/local/bin:/usr/bin:/bin:/usr/games:/usr/sbin:/sbin:/usr/pkg/sbin';

my $conffile = "${SYSCONFDIR}/mserv_irman.conf";

if (!getopts('c:hvV', \%opt) || $opt{h})
    {
    print "Usage: mserv_irman [opts]
opts:	-c file	Use config 'file' instead of $conffile.
	-h	This help
	-v	Verbose
	-V	Display version and exit
";
    exit;
    }

if ($opt{V})
    { print "$version\n"; exit; }

tie(%cfg, 'Config::IniFiles', ( -file => $opt{c} ?$opt{c} :$conffile ))
    || die "Unable to read config";

my $host = `hostname -s`;
chomp $host;

# Override sections on a per host basis
#
foreach my $section (qw(general irman remotes owners))
    {
    if ($cfg{"${section}_${host}"})
	{ $cfg{$section} = $cfg{"${section}_${host}"}; }
    }

$cfg{general}{default_debounce} ||= 0.5;
$cfg{general}{idle_timeout} ||= 3600;
$cfg{general}{flush_timeout} ||= 60;
$cfg{general}{favourite_inserts} ||= 5;
$cfg{general}{volume_up} ||= 1;
$cfg{general}{volume_down} ||= 1;

my $mserv = new MservCmd($cfg{irman});
my $die;

sub catch_signal
    {
    my $signame = shift;
    print "In catch_signal\n";
    $die = $signame;
    }

$SIG{HUP} = \&catch_signal;
$SIG{INT} = \&catch_signal;
$SIG{TERM} = \&catch_signal;

my @cmdlist = (
    { patt => '^mute$',   	fn => 'volume',  arg => '10', },
    { patt => '^(pause|10)$', 	fn => 'pause_toggle', },
    { patt => '^play$',  	fn => 'play', },
    { patt => '^stop$',  	fn => 'stop', 		     inactive => 1 },
    { patt => '^repeat$',	fn => 'repeat', },
    { patt => '^(next|fwd|ch\+)$', fn => 'next', },
    { patt => '^(prev|ch-)$',	fn => 'prev', },
    { patt => '^1$',		fn => 'rate',    arg => 'awful', },
    { patt => '^2$',		fn => 'rate',    arg => 'bad', },
    { patt => '^3$',		fn => 'rate',    arg => 'neutral',},
    { patt => '^4$',		fn => 'rate',    arg => 'good', },
    { patt => '^5$',		fn => 'rate',    arg => 'superb',},
    { patt => '^6$',		fn => 'queue_unrated' },
    { patt => '^7$',		fn => 'more',    arg => 'album' },
    { patt => '^8$',		fn => 'more',    arg => 'artist' },
    { patt => '^9$',		fn => 'queue_superb' },
    { patt => '^recall$',	fn => 'recall', },
    { patt => '^(0|cancel|clear|power|reset)$', fn => 'clear' },
    { patt => '^vol\+$',   	fn => 'volume', 
		    arg => "+$cfg{general}{volume_up}", debounce => 0.4 },
    { patt => '^vol-$',   	fn => 'volume', 
		    arg => "-$cfg{general}{volume_down}", debounce => 0.4 },
    # { patt => '^()$', 			     inactive => 1 },
    { patt => '.*',		fn => 'pause_toggle', },
    );

Owner::initlist($cfg{owners});
Remote::initlist($cfg{remotes});

$mserv->process();

exit;

sub verbose
    { $opt{v} && print @_; }

#------------------------------------------------------------------------------

package Owner;

my %owner_lookup;

# Special methods

sub initlist
    {
    my ($list) = @_;
    foreach my $user (keys %{$list})
	{ $owner_lookup{$user} = new Owner($user, $list->{$user}); }
    }

sub lookup
    { $owner_lookup{$_[0]}; }

sub new
    {
    my $class = shift;

    my $self = {};
    bless $self, $class;
    $self->user($_[0]);
    $self->pass($_[1]);
    return $self;
    }

# Ordinary methods

sub active
    {
    my $self = shift;
    if ($self->sock && $self->host ne $self->sockhost)
	{ $self->inactive('moved to another mserv'); }
    if (!$self->sock)
	{
	$self->sock(IO::Socket::INET->new(PeerAddr => $self->host));
	if (!$self->sock)
	    { return; }
	$self->sendcmd(); # Eat the welcome banner
	$self->sendcmd('user '.$self->user);
	$self->sendcmd('pass '.$self->pass.' computer');
	$self->sendcmd('emote has connected via mserv_irman...');
	$self->sockhost($self->host);
	}
    $self->last_active(time);
    }

sub flushall
    {
    foreach my $owner (values %owner_lookup)
	{
	if ($owner->sock)
	    {
	    $owner->sock->blocking(0);
	    my @dummy = $owner->sock->getlines();
	    $owner->sock->blocking(1);
	    if ($owner->last_active + $cfg{general}{idle_timeout} < time)
		{ $owner->inactive('idled out'); }
	    }
	}
    }

sub host
    {
    my $self = shift;
    @_ ? ($self->{_host} = $_[0]) : $self->{_host};
    }

sub inactive
    {
    my $self = shift;
    my($why) = @_;

    if ($self->sock)
	{
	main::verbose($self->user." $why\n");
	$self->sendcmd("emote has $why from mserv_irman...\n");
	close($self->sock);
	$self->sock(undef);
	}
    }

sub last_active
    {
    my $self = shift;
    @_ ? ($self->{_last_active} = $_[0]) : $self->{_last_active};
    }

sub sendcmd
    {
    my $self = shift;
    my($cmd, $errok) = @_;
    my(@output);
    if ($cmd)
	{
	main::verbose("$cmd\n");
	$self->sock->print("$cmd\n");
	}
    while($_ = $self->sock->getline())
	{
	s/\r\n$//;
	if ($_ eq '.')
	    { last; }
	push(@output, $_);
	}
    if (!@output)
	{ print STDERR "** No data returned for '$cmd'\n"; }
    elsif (!$errok && substr($output[0], 0, 1) eq '5')
	{ print STDERR "** ERROR returned for '$cmd'\n", @output; }
    @output;
    }

sub pass
    {
    my $self = shift;
    @_ ? ($self->{_pass} = $_[0]) : $self->{_pass};
    }

sub sock
    {
    my $self = shift;
    @_ ? ($self->{_sock} = $_[0]) : $self->{_sock};
    }

sub sockhost
    {
    my $self = shift;
    @_ ? ($self->{_sockhost} = $_[0]) : $self->{_sockhost};
    }

sub user
    {
    my $self = shift;
    @_ ? ($self->{_user} = $_[0]) : $self->{_user};
    }

1;

#------------------------------------------------------------------------------

package Mserv;

# Special methods

sub new
    {
    my $class = shift;

    my $self = {};
    bless $self, $class;

    my $host = $_[0];
    $host ||= 'localhost';
    (index($host, ':') == -1) && ($host .= ':4444');
    $self->host($host);

    return $self;
    }

#------------------------------------------------------------------------------

package Irman;

use Time::HiRes qw (time);

# Special methods

sub new
    {
    my $class = shift;

    my $self = {};
    bless $self, $class;
    $self->cmd($_[0]);
    my $host = $_[1];
    $host ||= 'localhost';
    $self->host($host);
    $self->mservcmd($_[2]);

    (index($host, ':') == -1) && ($host .= ':4444');

    my $fh;
    my $pid = open($fh, $self->cmd.'|');
    if (!$pid)
	{
	print STDERR "Unable to setup irman '".$self->cmd."'\n";
	return;
	}
    main::verbose("irman_name read: '".$self->cmd."'\n");
    $self->pid($pid);
    $self->fh($fh);
    $self->last_button('');
    $self->debounce(0);
    return $self;
    }

sub DESTROY
    {
    my $self = shift;
    if ($self->pid())
	{
	kill(1, $self->pid);
	delete $self->{_pid};
	}
    if ($self->fh())
	{
	close($self->fh);
	delete $self->{_fh};
	}
    }

# Ordinary methods

sub alive
    {
    my $self = shift;
    @_ ? ($self->{_alive} = $_[0]) : $self->{_alive};
    }

sub debounce
    {
    my $self = shift;
    @_ ? ($self->{_debounce} = $_[0]) : $self->{_debounce};
    }

sub fh
    {
    my $self = shift;
    @_ ? ($self->{_fh} = $_[0]) : $self->{_fh};
    }

sub host
    {
    my $self = shift;
    @_ ? ($self->{_host} = $_[0]) : $self->{_host};
    }

sub pid
    {
    my $self = shift;
    @_ ? ($self->{_pid} = $_[0]) : $self->{_pid};
    }

sub read_button
    {
    my $self = shift;
    my $line = $self->fh->getline();

    if (!$line) # irman failed - keep going if any left
	{
	print STDERR "irman '".$self->cmd."' died\n";
	return;
	}
    chomp $line;

    if ($line eq 'ok')
	{
	$self->alive(1);
	main::verbose("irman '".$self->cmd."' setup OK\n");
	return(0);
	}
    unless ($line =~ /^rx `(.+)'$/)
	{
	main::verbose("Unknown output from irman_name '$line'");
	return(0);
	}

    my $stamp = time;
    if ($self->last_button eq $1 && $stamp <
				$self->last_button_stamp + $self->debounce)
	{
	main::verbose(">>Debounced repeated $1 ".$self->last_button." $stamp ".$self->last_button_stamp."\n");
	return(0);
	}
    $self->last_button_stamp($stamp);
    $self->last_button($1);
    return($1);
    }

sub cmd
    {
    my $self = shift;
    @_ ? ($self->{_cmd} = $_[0]) : $self->{_cmd};
    }

sub last_button
    {
    my $self = shift;
    @_ ? ($self->{_last_button} = $_[0]) : $self->{_last_button};
    }

sub last_button_stamp
    {
    my $self = shift;
    @_ ? ($self->{_last_button_stamp} = $_[0]) : $self->{_last_button_stamp};
    }

sub mservcmd
    {
    my $self = shift;
    @_ ? ($self->{_mservcmd} = $_[0]) : $self->{_mservcmd};
    }

#------------------------------------------------------------------------------

package Remote;

my %remote_lookup;

# Special methods

sub initlist
    {
    my ($list) = @_;
    foreach my $name (keys %{$list})
	{
	$remote_lookup{$name} = new Remote($name, $list->{$name});
	}
    }

sub lookup
    { $remote_lookup{$_[0]}; }

sub new
    {
    my $class = shift;

    my $self = {};
    bless $self, $class;
    $self->name($_[0]);
    $self->ownerlist([split(' ', $_[1])]);
    foreach my $user (@{$self->ownerlist})
	{
	$self->owner(Owner::lookup($user)) && return $self;
	}
    return;
    }

# Ordinary methods

sub name
    {
    my $self = shift;
    @_ ? ($self->{_name} = $_[0]) : $self->{_name};
    }

sub owner
    {
    my $self = shift;
    @_ ? ($self->{_owner} = $_[0]) : $self->{_owner};
    }

sub ownerlist
    {
    my $self = shift;
    @_ ? ($self->{_ownerlist} = $_[0]) : $self->{_ownerlist};
    }

#------------------------------------------------------------------------------

package MservCmd;

sub cmd_output
    {
    my $self = shift;
    my($cmd) = @_;
    my @output;

    open(CMD, "$cmd|") || die("Unable to run '$cmd': $!");
    @output = <CMD>;
    close(CMD);
    @output;
    }

sub queue_superb
    {
    my $self = shift;

    $self->queuefilter('$user=superb', '!$user=bad&!$user=awful');
    }

sub queue_unrated
    {
    my $self = shift;
    my($tag) =
	'!$user=superb&!$user=good&!$user=neutral&!$user=bad&!$user=awful';
    $self->queuefilter($tag, $tag);
    }

sub queuefilter
    {
    my $self = shift;
    my($myfilter, $otherfilter) = @_;
    my $filter;

    $myfilter =~ s/\$user/$self->owner->user/ge;

    $filter = $myfilter;
    foreach my $user ($self->who)
	{
	($user eq $self->owner->user) && next;
	my $next = $otherfilter;
	$next =~ s/\$user/$user/g;
	$filter .= '&'. $next;
	}

    # If we cannot find enough tracks, try to fill remaining places with
    # 'otherfilter' ignored.
    my $pushed = $self->queuerandom($cfg{general}{favourite_inserts},
					    $self->tracks("searchf $filter"));
    if ($pushed < $cfg{general}{favourite_inserts})
	{
	$self->queuerandom($cfg{general}{favourite_inserts} - $pushed,
					    $self->tracks("searchf $myfilter"));
	}
    }

sub more
    {
    my $self = shift;
    my($tag) = @_;
    my $playing = $self->track_info();
    my (@tracklist, @tracklist_extra);

    $playing || return;
    foreach my $track (
	    ($tag eq 'artist')
		? $self->tracks("search $playing->{artist}")
		:($tag eq 'album')
		    ? $self->albumtracks("asearch $playing->{albumname}")
		    : ())
	{
	if ($tag eq 'artist' && $track->{artist} ne $playing->{artist})
	    { next; }
	elsif ($tag eq 'album' && $track->{albumname} ne $playing->{albumname})
	    { next; }
	if ($track->{rate} eq 'AWFUL' || $track->{rate} eq 'BAD' ||
						$track->{rate} eq 'NEUTRAL')
	    { next; }
	if ($track->{album} == $playing->{album} &&
				    $track->{track} == $playing->{track} )
	    { next; }
	if ($track->{rate} eq 'GOOD' || $track->{rate} eq 'SUPERB')
	    { push(@tracklist_extra, $track); }
	else
	    { push(@tracklist, $track); }
	}

    # Try unrated first, then GOOD/SUPERB
    $self->queuerandom($cfg{general}{favourite_inserts}, @tracklist);
    if (@tracklist < $cfg{general}{favourite_inserts})
	{
	$self->queuerandom($cfg{general}{favourite_inserts} - @tracklist,
							@tracklist_extra);
	}
    }

sub clear  { my $self = shift; $self->owner->sendcmd('clear'); }
sub next   { my $self = shift; $self->owner->sendcmd('next'); }
sub pause  { my $self = shift; $self->owner->sendcmd('pause'); }
sub play   { my $self = shift; $self->owner->sendcmd('play'); }
sub repeat { my $self = shift; $self->owner->sendcmd('repeat'); }
sub stop   { my $self = shift; $self->owner->sendcmd('stop'); }

sub process
    {
    my $self = shift;

    while(! $die)
	{
	Owner::flushall();
	my @ready = $self->rec->can_read($cfg{general}{flush_timeout});

	while (@ready)
	    {
	    $self->irman($self->fh2irman->{shift @ready});
	    my $button = $self->irman->read_button();
	    if (!defined $button)
		{
		if ($self->irman->alive())
		    { $self->irman_restart(); }
		else
		    { $self->irman_destroy(); }
		}
	    elsif ($button)
		{ $self->process_button($button); }
	    }
	}
    main::verbose("Exiting...\n");
    }

sub process_button
    {
    my $self = shift;
    my ($button) = @_;

    if ($button =~ /^(.+)-([^.-]+-?)(\.\d+|)$/)
	{
	unless ($self->remote($1))
	    {
	    print STDERR "Unknown remote '$1' ignored\n";
	    return;
	    }
	$button = $2;
	}
    else
	{
	print STDERR "Unparsable button - $button\n";
	return;
	}

    main::verbose("[$button.".time."] ");
    $button =~ s/\.\d+$//;
    foreach my $cmd (@cmdlist)
	{
	if ($button =~ /$cmd->{patt}/)
	    {
	    $self->irman->debounce($cmd->{debounce}
				?$cmd->{debounce}
				:$cfg{general}{default_debounce});
	    $self->owner->active();
	    my $fn = $cmd->{fn};
	    if ($fn)
		{ $self->$fn($cmd->{arg}); }
	    if ($cmd->{inactive})
		{ $self->owner->inactive("exited"); }
	    last;
	    }
	}
    }

sub pause_toggle
    {
    my $self = shift;
    my($output) = $self->owner->sendcmd('pause');
    if ($output =~ /^4/)
	{ $self->play(); }
    }

sub rate
    {
    my $self = shift;
    my ($rate) = @_;
    $self->owner->sendcmd("rate $rate");
    }

sub prev
    {
    my $self = shift;
    $self->recall();
    $self->next();
    }

sub irman
    {
    my $self = shift;
    @_ ? ($self->{_irman} = $_[0]) : $self->{_irman};
    }

sub irman_create
    {
    my $self = shift;
    my($irman_cmd, $irman_host) = @_;
    my $irman = new Irman ($irman_cmd, $irman_host, $self);
    if ($irman)
	{
	$self->rec->add($irman->fh);
	$self->{_fh2irman}{$irman->fh} = $irman;
	}
    }

sub irman_restart
    {
    my $self = shift;

    my $irman_cmd = $self->irman->cmd();
    my $irman_host = $self->irman->host();
    $self->irman_destroy();
    $self->irman_create($irman_cmd, $irman_host);
    }

sub irman_destroy
    {
    my $self = shift;
    $self->rec->remove($self->irman->fh);
    delete $self->{_fh2irman}{$self->irman->fh};
    delete $self->{_irman};
    }

sub rec
    {
    my $self = shift;
    @_ ? ($self->{_rec} = $_[0]) : $self->{_rec};
    }

sub fh2irman
    {
    my $self = shift;
    @_ ? ($self->{_fh2irman} = $_[0]) : $self->{_fh2irman};
    }

sub new
    {
    my $class = shift;
    my $self = {};
    bless $self, $class;

    srand(time());
    my $cfg = shift;

    $self->rec(new IO::Select);
    if (!$cfg || ! (keys %{$cfg}))
	{ $cfg = { 'irman_name' => 'localhost:4444' }; }
    foreach my $irman_cmd (keys %{$cfg})
	{ $self->irman_create($irman_cmd, $cfg->{$irman_cmd}); }
    if (! $self->{_fh2irman})
	{ die("No valid irman"); }

    return $self;
    }

sub queue
    {
    my $self = shift;
    my($ent) = @_;
    $self->owner->sendcmd("queue $ent->{album} $ent->{track}");
    }

sub queuerandom
    {
    my $self = shift;
    my ($number, @tracklist) = @_;
    my %queued;
    my $count = 0;
    foreach my $track ($self->tracks('queue'))
	{ $queued{"$track->{album}/$track->{track}"} = 1; }
    my @eligible;
    foreach my $track (@tracklist)
	{
	if ($queued{"$track->{album}/$track->{track}"}) # Already on queue
	    { next; }
	push(@eligible, $track);
	}
    for (my $loop = 0 ; @eligible && $loop < $number; ++$loop)
	{
	my $pick = rand(@eligible);
	$self->queue($eligible[$pick]);
	++$count;
	splice(@eligible, $pick, 1);
	}
    return $count;
    }

sub recall
    {
    my $self = shift;
    $self->queue(($self->tracks('history'))[1]);
    }

sub albumtracks
    {
    my $self = shift;
    my(@output) = $self->owner->sendcmd(@_);
    shift @output;
    my(@albums, @tracklist);
    foreach my $line (@output)
	{
	my(@val) = split('\t', $line);
	$val[0] =~ /(\d+)/ || die "Album '$val[0]' not a number";
	$val[0] = $1;
	push(@albums, $val[0]);
	}
    foreach my $album (@albums)
	{ push(@tracklist, $self->tracks("tracks $album", 1)); }
    @tracklist;
    }

sub tracks
    {
    my $self = shift;
    my($cmd, $albumline) = @_;
    my(@output) = $self->owner->sendcmd($cmd);
    shift @output;
    my(@tracks, $albumname);
    if ($albumline)
	{
	$albumname = (split('\t', shift @output))[2];
	$albumname =~ s:.*/::;
	}
    foreach my $line (@output)
	{
	my(@val) = split('\t', $line);
	if ($val[5] =~ /\d/)
	    { unshift(@val, 'unknown'); }
	push(@tracks, $self->trackclean(
		    { who => $val[0], album => $val[1], track => $val[2],
		     artist => $val[3], title => $val[4], rate => $val[5],
		     albumname => $albumname }));
	}
    @tracks;
    }

sub trackclean
    {
    my $self = shift;
    my ($track) = @_;
    $track->{album} =~ /(\d+)/ || die "Album '$track->{album}' not a number";
    $track->{album} = $1;
    $track->{track} =~ /(\d+)/ || die "Track '$track->{track}' not a number";
    $track->{track} = $1;
    if ($track->{artist} =~ /([\w ]+)/)
	{ $track->{artist} = $1; }
    else
	{ $track->{artist} = 'Invalid'; }
    if ($track->{albumname})
	{
	if ($track->{albumname} =~ /([\w ]+)/)
	    { $track->{albumname} = $1; }
	else
	    { $track->{albumname} = 'Invalid'; }
	}
    $track;
    }

sub track_info
    {
    my $self = shift;
    my(@info) = $self->owner->sendcmd('info');
    if (!@info)
	{ return; }
    my(@val) = split('\t', $info[1]);
    $val[3] =~ s#.*/##;
    $self->trackclean({ album => $val[0], track => $val[1],
		 albumname => $val[3],
		 artist => $val[4], title => $val[5], rating => $val[11] });
    }

sub owner
    {
    my $self = shift;
    $self->remote && $self->remote->owner;
    }

sub remote
    {
    my $self = shift;
    if (@_)
	{
	if ($self->{_remote} = Remote::lookup(@_))
	    { $self->remote->owner->host($self->irman->host); }
	}
    $self->{_remote};
    }

sub volume
    {
    my $self = shift;
    my($value) = @_;
    $self->owner->sendcmd("volume $value", 1);
    }

sub who
    {
    my $self = shift;
    my(@output) = $self->owner->sendcmd('who');
    my(%who);
    shift @output;
    foreach my $line (@output)
	{ ($line =~ /^([\w\d]+)\t/) && ($who{$1} = 1); }
    return (sort keys %who);
    }

1;
